@@ -30,7 +30,9 @@
 #include "errors.h"
 #include "keydb.h"
 #include "memory.h"
+#include "photoid.h"
 #include "util.h"
+#include "ttyio.h"
 #include "trustdb.h"
 #include "main.h"
 #include "i18n.h"
@@ -38,7 +40,6 @@
 static void list_all(int);
 static void list_one( STRLIST names, int secret);
 static void list_keyblock( KBNODE keyblock, int secret );
-static void fingerprint( PKT_public_key *pk, PKT_secret_key *sk );
 
 
 /****************
@@ -63,45 +64,116 @@ secret_key_list( STRLIST list )
 	list_one( list, 1 );
 }
 
+void
+show_policy_url(PKT_signature *sig,int indent)
+{
+  const byte *p;
+  size_t len;
+  int seq=0;
+
+  while((p=enum_sig_subpkt(sig->hashed,SIGSUBPKT_POLICY,&len,&seq)))
+    {
+      int i;
+
+      for(i=0;i<indent;i++)
+	putchar(' ');
+
+      /* This isn't UTF8 as it is a URL(?) */
+      printf(_("Signature policy: "));
+      print_string(stdout,p,len,0);
+      printf("\n");
+    }
+}
+
+void
+show_notation(PKT_signature *sig,int indent)
+{
+  const byte *p;
+  size_t len;
+  int seq=0;
+
+  /* There may be multiple notations in the same sig. */
+
+  while((p=enum_sig_subpkt(sig->hashed,SIGSUBPKT_NOTATION,&len,&seq)))
+    if(len>=8)
+      {
+	int n1,n2,i;
+
+	n1=(p[4]<<8)|p[5];
+	n2=(p[6]<<8)|p[7];
+
+	if(8+n1+n2!=len)
+	  {
+	    log_info(_("WARNING: invalid notation data found\n"));
+	    return;
+	  }
+
+	for(i=0;i<indent;i++)
+	  putchar(' ');
+
+	/* This is UTF8 */
+	printf(_("Signature notation: "));
+	print_utf8_string(stdout,p+8,n1);
+	printf("=");
+
+	if(*p&0x80)
+	  print_utf8_string(stdout,p+8+n1,n2);
+	else
+	  printf("[ %s ]",_("not human readable"));
+
+	printf("\n");
+      }
+  else
+    log_info(_("WARNING: invalid notation data found\n"));
+}
 
 static void
 list_all( int secret )
 {
-    KBPOS kbpos;
+    KEYDB_HANDLE hd;
     KBNODE keyblock = NULL;
     int rc=0;
-    int lastresno;
+    const char *lastresname, *resname;
 
-    rc = enum_keyblocks( secret? 5:0, &kbpos, &keyblock );
+    hd = keydb_new (secret);
+    if (!hd)
+        rc = G10ERR_GENERAL;
+    else
+        rc = keydb_search_first (hd);
     if( rc ) {
 	if( rc != -1 )
-	    log_error("enum_keyblocks(open) failed: %s\n", g10_errstr(rc) );
+	    log_error("keydb_search_first failed: %s\n", g10_errstr(rc) );
 	goto leave;
     }
 
-    lastresno = -1;
-    while( !(rc = enum_keyblocks( 1, &kbpos, &keyblock )) ) {
-	if( lastresno != kbpos.resno ) {
-	    const char *s = keyblock_resource_name( &kbpos );
+    lastresname = NULL;
+    do {
+        rc = keydb_get_keyblock (hd, &keyblock);
+        if (rc) {
+            log_error ("keydb_get_keyblock failed: %s\n", g10_errstr(rc));
+            goto leave;
+        }
+        resname = keydb_get_resource_name (hd);
+	if (lastresname != resname ) {
 	    int i;
 
-	    lastresno = kbpos.resno;
-	    printf("%s\n", s );
-	    for(i=strlen(s); i; i-- )
+	    printf("%s\n", resname );
+	    for(i=strlen(resname); i; i-- )
 		putchar('-');
 	    putchar('\n');
+            lastresname = resname;
 	}
         merge_keys_and_selfsig( keyblock );
 	list_keyblock( keyblock, secret );
-	release_kbnode( keyblock ); keyblock = NULL;
-    }
-
+	release_kbnode( keyblock ); 
+        keyblock = NULL;
+    } while (!(rc = keydb_search_next (hd)));
     if( rc && rc != -1 )
-	log_error("enum_keyblocks(read) failed: %s\n", g10_errstr(rc));
+	log_error ("keydb_search_next failed: %s\n", g10_errstr(rc));
 
   leave:
-    enum_keyblocks( 2, &kbpos, &keyblock ); /* close */
-    release_kbnode( keyblock );
+    release_kbnode (keyblock);
+    keydb_release (hd);
 }
 
 
@@ -111,7 +183,19 @@ list_one( STRLIST names, int secret )
     int rc = 0;
     KBNODE keyblock = NULL;
     GETKEY_CTX ctx;
+    const char *resname;
+    char *keyring_str = N_("Keyring");
+    int i;
 
+    /* fixme: using the bynames function has the disadvantage that we
+     * don't know wether one of the names given was not found.  OTOH,
+     * this function has the advantage to list the names in the
+     * sequence as defined by the keyDB and does not duplicate
+     * outputs.  A solution could be do test whether all given have
+     * been listed (this needs a way to use the keyDB search
+     * functions) or to have the search function return indicators for
+     * found names.  Yet another way is to use the keydb search
+     * facilities directly. */
     if( secret ) {
 	rc = get_seckey_bynames( &ctx, NULL, names, &keyblock );
 	if( rc ) {
@@ -120,6 +204,13 @@ list_one( STRLIST names, int secret )
 	    return;
 	}
 	do {
+	    if (opt.show_keyring) {
+		resname = keydb_get_resource_name (get_ctx_handle(ctx));
+		printf("%s: %s\n", keyring_str, resname);
+		for(i = strlen(resname) + strlen(keyring_str) + 2; i; i-- )
+		    putchar('-');
+		putchar('\n');
+	    }
 	    list_keyblock( keyblock, 1 );
 	    release_kbnode( keyblock );
 	} while( !get_seckey_next( ctx, NULL, &keyblock ) );
@@ -133,6 +224,13 @@ list_one( STRLIST names, int secret )
 	    return;
 	}
 	do {
+	    if (opt.show_keyring) {
+		resname = keydb_get_resource_name (get_ctx_handle(ctx));
+		printf("%s: %s\n", keyring_str, resname);
+		for(i = strlen(resname) + strlen(keyring_str) + 2; i; i-- )
+		    putchar('-');
+		putchar('\n');
+	    }
 	    list_keyblock( keyblock, 0 );
 	    release_kbnode( keyblock );
 	} while( !get_pubkey_next( ctx, NULL, &keyblock ) );
@@ -157,12 +255,12 @@ print_key_data( PKT_public_key *pk, u32 *keyid )
 static void
 print_capabilities (PKT_public_key *pk, PKT_secret_key *sk, KBNODE keyblock)
 {
-    unsigned int usage = pk? pk->pubkey_usage : sk->pubkey_usage;
+    unsigned int use = pk? pk->pubkey_usage : sk->pubkey_usage;
     
-    if ( usage & PUBKEY_USAGE_ENC ) {
+    if ( use & PUBKEY_USAGE_ENC ) {
         putchar ('e');
     }
-    if ( usage & PUBKEY_USAGE_SIG ) {
+    if ( use & PUBKEY_USAGE_SIG ) {
         putchar ('s');
         putchar ('c');
     }
@@ -243,6 +341,12 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 
     for( kbctx=NULL; (node=walk_kbnode( keyblock, &kbctx, 0)) ; ) {
 	if( node->pkt->pkttype == PKT_USER_ID && !opt.fast_list_mode ) {
+            /* don't list revoked UIDS unless we are in verbose mode and 
+             * signature listing has not been requested */
+            if ( !opt.verbose && !opt.list_sigs
+                 && node->pkt->pkt.user_id->is_revoked )
+                continue; 
+
 	    if( any ) 
                 printf("uid%*s", 28, "");
 
@@ -253,11 +357,15 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 	    putchar('\n');
 	    if( !any ) {
 		if( opt.fingerprint )
-		    fingerprint( pk, sk );
+		    print_fingerprint( pk, sk, 0 );
 		if( opt.with_key_data )
 		    print_key_data( pk, keyid );
 		any = 1;
 	    }
+
+	    if(opt.show_photos && node->pkt->pkt.user_id->attribs!=NULL &&
+	       node->pkt->pkt.user_id->attribs->type==ATTRIB_JPEG)
+	      show_photo(node->pkt->pkt.user_id->attribs,pk);
 	}
 	else if( node->pkt->pkttype == PKT_PUBLIC_SUBKEY ) {
 	    u32 keyid2[2];
@@ -266,7 +374,7 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 	    if( !any ) {
 		putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk ); /* of the main key */
+		    print_fingerprint( pk, sk, 0 ); /* of the main key */
 		any = 1;
 	    }
 
@@ -280,7 +388,7 @@ list_keyblock_print ( KBNODE keyblock, int secret )
             }
             putchar('\n');
 	    if( opt.fingerprint > 1 )
-		fingerprint( pk2, NULL );
+		print_fingerprint( pk2, NULL, 0 );
 	    if( opt.with_key_data )
 		print_key_data( pk2, keyid2 );
 	}
@@ -291,7 +399,7 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 	    if( !any ) {
 		putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk ); /* of the main key */
+		    print_fingerprint( pk, sk, 0 ); /* of the main key */
 		any = 1;
 	    }
 
@@ -301,7 +409,7 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 					   (ulong)keyid2[1],
 					   datestr_from_sk( sk2 ) );
 	    if( opt.fingerprint > 1 )
-		fingerprint( NULL, sk2 );
+		print_fingerprint( NULL, sk2, 0 );
 	}
 	else if( opt.list_sigs && node->pkt->pkttype == PKT_SIGNATURE ) {
 	    PKT_signature *sig = node->pkt->pkt.signature;
@@ -309,7 +417,11 @@ list_keyblock_print ( KBNODE keyblock, int secret )
             char *sigstr;
 
 	    if( !any ) { /* no user id, (maybe a revocation follows)*/
-		if( sig->sig_class == 0x20 )
+	      /* Check if the pk is really revoked - there could be a
+                 0x20 sig packet there even if we are not revoked
+                 (say, if a revocation key issued the packet, but the
+                 revocation key isn't present to verify it.) */
+		if( sig->sig_class == 0x20 && pk->is_revoked )
 		    puts("[revoked]");
 		else if( sig->sig_class == 0x18 )
 		    puts("[key binding]");
@@ -318,7 +430,7 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 		else
 		    putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk );
+		    print_fingerprint( pk, sk, 0 );
 		any=1;
 	    }
 
@@ -329,6 +441,8 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 	       sigstr = "sig";
 	    else if( sig->sig_class == 0x18 )
 	       sigstr = "sig";
+	    else if( sig->sig_class == 0x1F )
+	       sigstr = "sig";
 	    else {
                 printf("sig                             "
 		       "[unexpected signature class 0x%02x]\n",sig->sig_class );
@@ -349,9 +463,17 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 		rc = 0;
 		sigrc = ' ';
 	    }
+
             fputs( sigstr, stdout );
-            printf("%c       %08lX %s  ",
-		    sigrc, (ulong)sig->keyid[1], datestr_from_sig(sig));
+	    printf("%c%c %c%c%c%c%c %08lX %s   ",
+                   sigrc,(sig->sig_class-0x10>0 &&
+                          sig->sig_class-0x10<4)?'0'+sig->sig_class-0x10:' ',
+                   sig->flags.exportable?' ':'L',
+                   sig->flags.revocable?' ':'R',
+                   sig->flags.policy_url?'P':' ',
+                   sig->flags.notation?'N':' ',
+                   sig->flags.expired?'X':' ',
+                   (ulong)sig->keyid[1], datestr_from_sig(sig));
 	    if( sigrc == '%' )
 		printf("[%s] ", g10_errstr(rc) );
 	    else if( sigrc == '?' )
@@ -363,6 +485,13 @@ list_keyblock_print ( KBNODE keyblock, int secret )
 		m_free(p);
 	    }
 	    putchar('\n');
+
+	    if(sig->flags.policy_url && opt.show_policy_url)
+	      show_policy_url(sig,3);
+
+	    if(sig->flags.notation && opt.show_notation)
+	      show_notation(sig,3);
+
 	    /* fixme: check or list other sigs here */
 	}
     }
@@ -399,8 +528,8 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 		    nbits_from_sk( sk ),
 		    sk->pubkey_algo,
 		    (ulong)keyid[0],(ulong)keyid[1],
-		    datestr_from_sk( sk ),
-		    sk->expiredate? strtimestamp(sk->expiredate):""
+		    colon_datestr_from_sk( sk ),
+		    colon_strtime (sk->expiredate)
 		    /* fixme: add LID here */ );
     }
     else {
@@ -418,7 +547,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
         else if ( opt.fast_list_mode || opt.no_expensive_trust_checks ) 
             ;
         else {
-            trustletter = query_trust_info( pk, NULL );
+            trustletter = get_validity_info ( pk, NULL );
             if( trustletter == 'u' )
                 ulti_hack = 1;
             putchar(trustletter);
@@ -427,14 +556,13 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 		    nbits_from_pk( pk ),
 		    pk->pubkey_algo,
 		    (ulong)keyid[0],(ulong)keyid[1],
-		    datestr_from_pk( pk ),
-		    pk->expiredate? strtimestamp(pk->expiredate):"" );
+		    colon_datestr_from_pk( pk ),
+		    colon_strtime (pk->expiredate) );
         if( pk->local_id )
             printf("%lu", pk->local_id );
         putchar(':');
-        if( pk->local_id && !opt.fast_list_mode
-            && !opt.no_expensive_trust_checks  )
-            putchar( get_ownertrust_info( pk->local_id ) );
+        if( !opt.fast_list_mode && !opt.no_expensive_trust_checks  )
+            putchar( get_ownertrust_info(pk) );
 	    putchar(':');
     }
     
@@ -445,7 +573,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
         print_capabilities (pk, sk, keyblock);
         putchar('\n');
         if( opt.fingerprint )
-            fingerprint( pk, sk );
+            print_fingerprint( pk, sk, 0 );
         if( opt.with_key_data )
             print_key_data( pk, keyid );
         any = 1;
@@ -467,15 +595,15 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 		    byte namehash[20];
 
 		    if( pk && !ulti_hack ) {
-			if( node->pkt->pkt.user_id->photo )
+			if( node->pkt->pkt.user_id->attrib_data )
 			    rmd160_hash_buffer( namehash,
-					    node->pkt->pkt.user_id->photo,
-					    node->pkt->pkt.user_id->photolen);
+					   node->pkt->pkt.user_id->attrib_data,
+					   node->pkt->pkt.user_id->attrib_len);
 			else
 			    rmd160_hash_buffer( namehash,
 					    node->pkt->pkt.user_id->name,
 					    node->pkt->pkt.user_id->len  );
-			trustletter = query_trust_info( pk, namehash );
+			trustletter = get_validity_info( pk, namehash );
 		    }
 		    else
 			trustletter = 'u';
@@ -492,7 +620,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
                 print_capabilities (pk, sk, keyblock);
                 putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk );
+		    print_fingerprint( pk, sk, 0 );
 		if( opt.with_key_data )
 		    print_key_data( pk, keyid );
 		any = 1;
@@ -508,7 +636,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
                 print_capabilities (pk, sk, keyblock);
                 putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk ); /* of the main key */
+		    print_fingerprint( pk, sk, 0 ); /* of the main key */
 		any = 1;
 	    }
 
@@ -529,8 +657,8 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 			nbits_from_pk( pk2 ),
 			pk2->pubkey_algo,
 			(ulong)keyid2[0],(ulong)keyid2[1],
-			datestr_from_pk( pk2 ),
-			pk2->expiredate? strtimestamp(pk2->expiredate):""
+			colon_datestr_from_pk( pk2 ),
+			colon_strtime (pk2->expiredate)
 			/* fixme: add LID and ownertrust here */
 						);
             if( pk->local_id ) /* use the local_id of the main key??? */
@@ -542,7 +670,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
             print_capabilities (pk2, NULL, NULL);
             putchar('\n');
 	    if( opt.fingerprint > 1 )
-		fingerprint( pk2, NULL );
+		print_fingerprint( pk2, NULL, 0 );
 	    if( opt.with_key_data )
 		print_key_data( pk2, keyid2 );
 	}
@@ -556,7 +684,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
                 print_capabilities (pk, sk, keyblock);
 		putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk ); /* of the main key */
+		    print_fingerprint( pk, sk, 0 ); /* of the main key */
 		any = 1;
 	    }
 
@@ -565,13 +693,13 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 			nbits_from_sk( sk2 ),
 			sk2->pubkey_algo,
 			(ulong)keyid2[0],(ulong)keyid2[1],
-			datestr_from_sk( sk2 ),
-			sk2->expiredate? strtimestamp(sk2->expiredate):""
+			colon_datestr_from_sk( sk2 ),
+			colon_strtime (sk2->expiredate)
                    /* fixme: add LID */ );
             print_capabilities (NULL, sk2, NULL);
             putchar ('\n');
 	    if( opt.fingerprint > 1 )
-		fingerprint( NULL, sk2 );
+		print_fingerprint( NULL, sk2, 0 );
 	}
 	else if( opt.list_sigs && node->pkt->pkttype == PKT_SIGNATURE ) {
 	    PKT_signature *sig = node->pkt->pkt.signature;
@@ -591,7 +719,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
                 print_capabilities (pk, sk, keyblock);
                 putchar('\n');
 		if( opt.fingerprint )
-		    fingerprint( pk, sk );
+		    print_fingerprint( pk, sk, 0 );
 		any=1;
 	    }
 
@@ -603,7 +731,8 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 	    else if( sig->sig_class == 0x18 )
 	       sigstr = "sig";
 	    else {
-                printf("sig::::::::::%02x:\n",sig->sig_class );
+                printf ("sig::::::::::%02x%c:\n",
+                        sig->sig_class, sig->flags.exportable?'x':'l');
 		continue;
 	    }
 	    if( opt.check_sigs ) {
@@ -625,9 +754,10 @@ list_keyblock_colon( KBNODE keyblock, int secret )
             putchar(':');
             if( sigrc != ' ' )
                 putchar(sigrc);
-            printf("::%d:%08lX%08lX:%s::::", sig->pubkey_algo,
+            printf("::%d:%08lX%08lX:%s:%s:::", sig->pubkey_algo,
 						 (ulong)sig->keyid[0],
-			   (ulong)sig->keyid[1], datestr_from_sig(sig));
+			   (ulong)sig->keyid[1], colon_datestr_from_sig(sig),
+		           colon_expirestr_from_sig(sig));
 	    if( sigrc == '%' )
 		printf("[%s] ", g10_errstr(rc) );
 	    else if( sigrc == '?' )
@@ -638,7 +768,7 @@ list_keyblock_colon( KBNODE keyblock, int secret )
                 print_string( stdout, p, n, ':' );
 		m_free(p);
 	    }
-            printf(":%02x:\n", sig->sig_class );
+            printf(":%02x%c:\n", sig->sig_class,sig->flags.exportable?'x':'l');
 	    /* fixme: check or list other sigs here */
 	}
     }
@@ -651,9 +781,9 @@ list_keyblock_colon( KBNODE keyblock, int secret )
 }
 
 /*
- * Reorder the keyblock so that the primary user ID comes first.
- * Fixme: Replace this by a generic sort function.
- */
+ * Reorder the keyblock so that the primary user ID (and not attribute
+ * packet) comes first.  Fixme: Replace this by a generic sort
+ * function.  */
 static void
 reorder_keyblock (KBNODE keyblock)
 {
@@ -661,12 +791,12 @@ reorder_keyblock (KBNODE keyblock)
     KBNODE last, node;
 
     for (node=keyblock; node; primary0=node, node = node->next) {
-	if( node->pkt->pkttype == PKT_USER_ID 
-            && node->pkt->pkt.user_id->is_primary ) {
+	if( node->pkt->pkttype == PKT_USER_ID &&
+	    !node->pkt->pkt.user_id->attrib_data &&
+            node->pkt->pkt.user_id->is_primary ) {
             primary = primary2 = node;
             for (node=node->next; node; primary2=node, node = node->next ) {
                 if( node->pkt->pkttype == PKT_USER_ID 
-                    || node->pkt->pkttype == PKT_PHOTO_ID 
                     || node->pkt->pkttype == PKT_PUBLIC_SUBKEY 
                     || node->pkt->pkttype == PKT_SECRET_SUBKEY ) {
                     break;
@@ -696,49 +826,93 @@ reorder_keyblock (KBNODE keyblock)
 static void
 list_keyblock( KBNODE keyblock, int secret )
 {
-    log_debug ("before reorder:\n");
-    dump_kbnode (keyblock);
     reorder_keyblock (keyblock);
-    log_debug ("after reorder:\n");
-    dump_kbnode (keyblock);
     if (opt.with_colons)
         list_keyblock_colon (keyblock, secret );
     else
         list_keyblock_print (keyblock, secret );
 }
 
-static void
-fingerprint( PKT_public_key *pk, PKT_secret_key *sk )
+/*
+ * standard function to print the finperprint.
+ * mode 0: as used in key listings, opt.with_colons is honored
+ *      1: print using log_info ()
+ *      2: direct use of tty
+ */
+void
+print_fingerprint (PKT_public_key *pk, PKT_secret_key *sk, int mode )
 {
-    byte *array, *p;
+    byte array[MAX_FINGERPRINT_LEN], *p;
     size_t i, n;
+    FILE *fp;
+    const char *text;
 
-    p = array = pk? fingerprint_from_pk( pk, NULL, &n )
-		   : fingerprint_from_sk( sk, NULL, &n );
-    if( opt.with_colons ) {
-	printf("fpr:::::::::");
-	for(i=0; i < n ; i++, p++ )
-	    printf("%02X", *p );
-	putchar(':');
+    if (mode == 1) {
+        fp = log_stream ();
+        text = _("Fingerprint:");
+    }
+    else if (mode == 2) {
+        fp = NULL; /* use tty */
+        /* Translators: this should fit into 24 bytes to that the fingerprint
+         * data is properly aligned with the user ID */
+        text = _("             Fingerprint:");
     }
     else {
-	printf("     Key fingerprint =");
-	if( n == 20 ) {
-	    for(i=0; i < n ; i++, i++, p += 2 ) {
-		if( i == 10 )
-		    putchar(' ');
-		printf(" %02X%02X", *p, p[1] );
+        fp = stdout;
+        text = _("     Key fingerprint =");
+    }
+  
+    if (sk)
+	fingerprint_from_sk (sk, array, &n);
+    else
+	fingerprint_from_pk (pk, array, &n);
+    p = array;
+    if (opt.with_colons && !mode) {
+	fprintf (fp, "fpr:::::::::");
+	for (i=0; i < n ; i++, p++ )
+	    fprintf (fp, "%02X", *p );
+	putc(':', fp);
+    }
+    else {
+        if (fp)
+            fputs (text, fp);
+        else
+            tty_printf ("%s", text);
+	if (n == 20) {
+	    for (i=0; i < n ; i++, i++, p += 2 ) {
+                if (fp) {
+                    if (i == 10 )
+                        putc(' ', fp);
+                    fprintf (fp, " %02X%02X", *p, p[1] );
+                }
+                else {
+                    if (i == 10 )
+                        tty_printf (" ");
+                    tty_printf (" %02X%02X", *p, p[1]);
+                }
 	    }
 	}
 	else {
-	    for(i=0; i < n ; i++, p++ ) {
-		if( i && !(i%8) )
-		    putchar(' ');
-		printf(" %02X", *p );
+	    for (i=0; i < n ; i++, p++ ) {
+                if (fp) {
+                    if (i && !(i%8) )
+                        putc (' ', fp);
+                    fprintf (fp, " %02X", *p );
+                }
+                else {
+                    if (i && !(i%8) )
+                        tty_printf (" ");
+                    tty_printf (" %02X", *p );
+                }
 	    }
 	}
     }
-    putchar('\n');
-    m_free(array);
+    if (fp)
+        putc ('\n', fp);
+    else
+        tty_printf ("\n");
 }
 
+
+
+
