@@ -37,7 +37,7 @@
 #include "status.h"
 #include "i18n.h"
 #include "trustdb.h"
-#include "hkp.h"
+#include "keyserver-internal.h"
 
 
 struct kidlist_item {
@@ -143,8 +143,6 @@ add_gpg_control( CTX c, PACKET *pkt )
     }   
     else if ( pkt->pkt.gpg_control->control == CTRLPKT_PIPEMODE ) {
         /* Pipemode control packet */
-#warning the --pipemode does not yet work 
-        /* FIXME: We have to do more sanity checks all over the place */
         if ( pkt->pkt.gpg_control->datalen < 2 ) 
             log_fatal ("invalid pipemode control packet length\n");
         if (pkt->pkt.gpg_control->data[0] == 1) {
@@ -219,11 +217,11 @@ add_signature( CTX c, PACKET *pkt )
 
     if( pkt->pkttype == PKT_SIGNATURE && !c->list ) {
 	/* This is the first signature for the following datafile.
-	 * G10 does not write such packets; instead it always uses
+	 * GPG does not write such packets; instead it always uses
 	 * onepass-sig packets.  The drawback of PGP's method
 	 * of prepending the signature to the data is
 	 * that it is not possible to make a signature from data read
-	 * from stdin.	(G10 is able to read PGP stuff anyway.) */
+	 * from stdin.	(GPG is able to read PGP stuff anyway.) */
 	node = new_kbnode( pkt );
 	c->list = node;
 	return 1;
@@ -249,8 +247,19 @@ proc_symkey_enc( CTX c, PACKET *pkt )
     if (enc->seskeylen)
 	log_error ("symkey_enc packet with session keys are not supported!\n");
     else {
+        int algo = enc->cipher_algo;
+	const char *s;
+
+	s = cipher_algo_to_string (algo);
+	if( s )
+	    log_info(_("%s encrypted data\n"), s );
+	else
+	    log_info(_("encrypted with unknown algorithm %d\n"), algo );
+
 	c->last_was_session_key = 2;
-	c->dek = passphrase_to_dek( NULL, 0, enc->cipher_algo, &enc->s2k, 0 );
+	c->dek = passphrase_to_dek( NULL, 0, algo, &enc->s2k, 0, NULL );
+        if (c->dek)
+            c->dek->algo_info_printed = 1;
     }
     free_packet(pkt);
 }
@@ -278,10 +287,10 @@ proc_pubkey_enc( CTX c, PACKET *pkt )
     }
 
     if( !opt.list_only && opt.override_session_key ) {
-	/* It does not make nuch sense to store the session key in
+	/* It does not make much sense to store the session key in
 	 * secure memory because it has already been passed on the
 	 * command line and the GCHQ knows about it */
-	c->dek = m_alloc( sizeof *c->dek );
+	c->dek = m_alloc_clear( sizeof *c->dek );
 	result = get_override_session_key ( c->dek, opt.override_session_key );
 	if ( result ) {
 	    m_free(c->dek); c->dek = NULL;
@@ -291,11 +300,12 @@ proc_pubkey_enc( CTX c, PACKET *pkt )
 	|| enc->pubkey_algo == PUBKEY_ALGO_DSA
 	|| is_RSA(enc->pubkey_algo)  ) {
 	if ( !c->dek && ((!enc->keyid[0] && !enc->keyid[1])
+                          || opt.try_all_secrets
 			  || !seckey_available( enc->keyid )) ) {
 	    if( opt.list_only )
 		result = -1;
 	    else {
-		c->dek = m_alloc_secure( sizeof *c->dek );
+		c->dek = m_alloc_secure_clear( sizeof *c->dek );
 		if( (result = get_session_key( enc, c->dek )) ) {
 		    /* error: delete the DEK */
 		    m_free(c->dek); c->dek = NULL;
@@ -370,7 +380,7 @@ print_pkenc_list( struct kidlist_item *list, int failed )
 		       strtimestamp(pk->timestamp) );
 	    fputs("      \"", log_stream() );
 	    p = get_user_id( list->kid, &n );
-	    print_string( log_stream(), p, n, '"' );
+	    print_utf8_string2 ( log_stream(), p, n, '"' );
 	    m_free(p);
 	    fputs("\"\n", log_stream() );
 	}
@@ -400,8 +410,10 @@ proc_encrypted( CTX c, PACKET *pkt )
 {
     int result = 0;
 
-    print_pkenc_list( c->pkenc_list, 1 );
-    print_pkenc_list( c->pkenc_list, 0 );
+    if (!opt.quiet) {
+        print_pkenc_list ( c->pkenc_list, 1 );
+        print_pkenc_list ( c->pkenc_list, 0 );
+    }
 
     write_status( STATUS_BEGIN_DECRYPTION );
 
@@ -409,14 +421,36 @@ proc_encrypted( CTX c, PACKET *pkt )
     if( opt.list_only )
 	result = -1;
     else if( !c->dek && !c->last_was_session_key ) {
-        int def_algo;
-	/* assume this is old conventional encrypted data
-	 * We use IDEA here if it is installed */
-        def_algo = check_cipher_algo (CIPHER_ALGO_IDEA)?
-                                      DEFAULT_CIPHER_ALGO : CIPHER_ALGO_IDEA;
-	c->dek = passphrase_to_dek( NULL, 0,
-		    opt.def_cipher_algo ? opt.def_cipher_algo
-					: DEFAULT_CIPHER_ALGO, NULL, 0 );
+        int algo;
+        STRING2KEY s2kbuf, *s2k = NULL;
+
+	/* assume this is old style conventional encrypted data */
+        if ( (algo = opt.def_cipher_algo))
+            log_info (_("assuming %s encrypted data\n"),
+                        cipher_algo_to_string(algo));
+        else if ( check_cipher_algo(CIPHER_ALGO_IDEA) ) {
+            algo = opt.def_cipher_algo;
+            if (!algo)
+                algo = opt.s2k_cipher_algo;
+	    idea_cipher_warn(1);
+            log_info (_("IDEA cipher unavailable, "
+                        "optimistically attempting to use %s instead\n"),
+                       cipher_algo_to_string(algo));
+        }
+        else {
+            algo = CIPHER_ALGO_IDEA;
+            if (!opt.def_digest_algo) {
+                /* If no digest is given we assume MD5 */
+                s2kbuf.mode = 0;
+                s2kbuf.hash_algo = DIGEST_ALGO_MD5;
+                s2k = &s2kbuf;
+            }
+            log_info (_("assuming %s encrypted data\n"), "IDEA");
+        }
+
+	c->dek = passphrase_to_dek ( NULL, 0, algo, s2k, 0, NULL );
+        if (c->dek)
+            c->dek->algo_info_printed = 1;
     }
     else if( !c->dek )
 	result = G10ERR_NO_SECKEY;
@@ -522,16 +556,29 @@ proc_plaintext( CTX c, PACKET *pkt )
 	if ( c->mfx.md2  )
 	    md_start_debug( c->mfx.md2, "verify2" );
     }
-    rc = handle_plaintext( pt, &c->mfx, c->sigs_only, clearsig );
-    if( rc == G10ERR_CREATE_FILE && !c->sigs_only) {
-	/* can't write output but we hash it anyway to
-	 * check the signature */
-	rc = handle_plaintext( pt, &c->mfx, 1, clearsig );
+    if ( c->pipemode.op == 'B' )
+        rc = handle_plaintext( pt, &c->mfx, 1, 0 );
+    else {
+        rc = handle_plaintext( pt, &c->mfx, c->sigs_only, clearsig );
+        if( rc == G10ERR_CREATE_FILE && !c->sigs_only) {
+            /* can't write output but we hash it anyway to
+             * check the signature */
+            rc = handle_plaintext( pt, &c->mfx, 1, clearsig );
+        }
     }
     if( rc )
 	log_error( "handle plaintext failed: %s\n", g10_errstr(rc));
     free_packet(pkt);
     c->last_was_session_key = 0;
+
+    /* We add a marker control packet instead of the plaintext packet.
+     * This is so that we can later detect invalid packet sequences.
+     */
+    n = new_kbnode (create_gpg_control (CTRLPKT_PLAINTEXT_MARK, NULL, 0));
+    if (c->list)
+        add_kbnode (c->list, n);
+    else 
+        c->list = n;
 }
 
 
@@ -572,11 +619,15 @@ proc_compressed( CTX c, PACKET *pkt )
  * Returns: 0 = valid signature or an error code
  */
 static int
-do_check_sig( CTX c, KBNODE node, int *is_selfsig )
+do_check_sig( CTX c, KBNODE node, int *is_selfsig, int *is_expkey )
 {
     PKT_signature *sig;
     MD_HANDLE md = NULL, md2 = NULL;
-    int algo, rc;
+    int algo, rc, dum2;
+    u32 dummy;
+
+    if(!is_expkey)
+      is_expkey=&dum2;
 
     assert( node->pkt->pkttype == PKT_SIGNATURE );
     if( is_selfsig )
@@ -609,8 +660,10 @@ do_check_sig( CTX c, KBNODE node, int *is_selfsig )
     }
     else if( (sig->sig_class&~3) == 0x10
 	     || sig->sig_class == 0x18
+             || sig->sig_class == 0x1f
 	     || sig->sig_class == 0x20
-	     || sig->sig_class == 0x30	) { /* classes 0x10..0x17,0x20,0x30 */
+	     || sig->sig_class == 0x28
+	     || sig->sig_class == 0x30	) { 
 	if( c->list->pkt->pkttype == PKT_PUBLIC_KEY
 	    || c->list->pkt->pkttype == PKT_PUBLIC_SUBKEY ) {
 	    return check_key_signature( c->list, node, is_selfsig );
@@ -628,9 +681,9 @@ do_check_sig( CTX c, KBNODE node, int *is_selfsig )
     }
     else
 	return G10ERR_SIG_CLASS;
-    rc = signature_check( sig, md );
+    rc = signature_check2( sig, md, &dummy, is_expkey );
     if( rc == G10ERR_BAD_SIGN && md2 )
-	rc = signature_check( sig, md2 );
+	rc = signature_check2( sig, md2, &dummy, is_expkey );
     md_close(md);
     md_close(md2);
 
@@ -656,43 +709,6 @@ print_userid( PACKET *pkt )
 }
 
 
-static void
-print_fingerprint( PKT_public_key *pk, PKT_secret_key *sk )
-{
-    byte array[MAX_FINGERPRINT_LEN], *p;
-    size_t i, n;
-
-    if( sk )
-	fingerprint_from_sk( sk, array, &n );
-    else
-	fingerprint_from_pk( pk, array, &n );
-    p = array;
-    if( opt.with_colons ) {
-	printf("fpr:::::::::");
-	for(i=0; i < n ; i++, p++ )
-	    printf("%02X", *p );
-	putchar(':');
-    }
-    else {
-	printf("     Key fingerprint =");
-	if( n == 20 ) {
-	    for(i=0; i < n ; i++, i++, p += 2 ) {
-		if( i == 10 )
-		    putchar(' ');
-		printf(" %02X%02X", *p, p[1] );
-	    }
-	}
-	else {
-	    for(i=0; i < n ; i++, p++ ) {
-		if( i && !(i%8) )
-		    putchar(' ');
-		printf(" %02X", *p );
-	    }
-	}
-    }
-    putchar('\n');
-}
-
 static void
 print_notation_data( PKT_signature *sig )
 {
@@ -700,8 +716,7 @@ print_notation_data( PKT_signature *sig )
     const byte *p;
     int seq = 0;
 
-    while( (p = enum_sig_subpkt( sig->hashed_data, SIGSUBPKT_NOTATION,
-				 &n, &seq )) ) {
+    while( (p = enum_sig_subpkt (sig->hashed, SIGSUBPKT_NOTATION, &n, &seq))) {
 	if( n < 8 ) {
 	    log_info(_("WARNING: invalid notation data found\n"));
 	    return;
@@ -721,16 +736,19 @@ print_notation_data( PKT_signature *sig )
 	print_string( log_stream(), p+n1, n2, 0 );
 	putc( '\n', log_stream() );
         write_status_buffer ( STATUS_NOTATION_NAME, p   , n1, 0 );
-        write_status_buffer ( STATUS_NOTATION_DATA, p+n1, n2, 20 );
+        write_status_buffer ( STATUS_NOTATION_DATA, p+n1, n2, 50 );
     }
-    if( (p = parse_sig_subpkt( sig->hashed_data, SIGSUBPKT_POLICY, &n ) )) {
+
+    seq=0;
+
+    while( (p = enum_sig_subpkt (sig->hashed, SIGSUBPKT_POLICY, &n, &seq) )) {
 	log_info(_("Policy: ") );
 	print_string( log_stream(), p, n, 0 );
 	putc( '\n', log_stream() );
         write_status_buffer ( STATUS_POLICY_URL, p, n, 0 );
     }
 
-    /* Now check wheter the key of this signature has some
+    /* Now check whether the key of this signature has some
      * notation data */
 
     /* TODO */
@@ -759,7 +777,7 @@ list_node( CTX c, KBNODE node )
 	    if( mainkey ) {
 		c->local_id = pk->local_id;
 		c->trustletter = opt.fast_list_mode?
-					   0 : query_trust_info( pk, NULL );
+					   0 : get_validity_info( pk, NULL );
 	    }
 	    printf("%s:", mainkey? "pub":"sub" );
 	    if( c->trustletter )
@@ -768,18 +786,18 @@ list_node( CTX c, KBNODE node )
 		    nbits_from_pk( pk ),
 		    pk->pubkey_algo,
 		    (ulong)keyid[0],(ulong)keyid[1],
-		    datestr_from_pk( pk ),
-		    pk->expiredate? strtimestamp(pk->expiredate):"" );
+		    colon_datestr_from_pk( pk ),
+		    colon_strtime (pk->expiredate) );
 	    if( c->local_id )
 		printf("%lu", c->local_id );
 	    putchar(':');
-	    if( c->local_id && !opt.fast_list_mode )
-		putchar( get_ownertrust_info( c->local_id ) );
+	    if( mainkey && !opt.fast_list_mode )
+                 putchar( get_ownertrust_info (pk) );
 	    putchar(':');
 	    if( node->next && node->next->pkt->pkttype == PKT_RING_TRUST) {
 		putchar('\n'); any=1;
 		if( opt.fingerprint )
-		    print_fingerprint( pk, NULL );
+		    print_fingerprint( pk, NULL, 0 );
 		printf("rtv:1:%u:\n",
 			    node->next->pkt->pkt.ring_trust->trustval );
 	    }
@@ -817,7 +835,7 @@ list_node( CTX c, KBNODE node )
 			putchar(':');
 		    putchar('\n');
 		    if( opt.fingerprint && !any )
-			print_fingerprint( pk, NULL );
+			print_fingerprint( pk, NULL, 0 );
 		    if( node->next
 			&& node->next->pkt->pkttype == PKT_RING_TRUST ) {
 			printf("rtv:2:%u:\n",
@@ -841,7 +859,7 @@ list_node( CTX c, KBNODE node )
 	if( !any )
 	    putchar('\n');
 	if( !mainkey && opt.fingerprint > 1 )
-	    print_fingerprint( pk, NULL );
+	    print_fingerprint( pk, NULL, 0 );
     }
     else if( (mainkey = (node->pkt->pkttype == PKT_SECRET_KEY) )
 	     || node->pkt->pkttype == PKT_SECRET_SUBKEY ) {
@@ -855,8 +873,8 @@ list_node( CTX c, KBNODE node )
 		    nbits_from_sk( sk ),
 		    sk->pubkey_algo,
 		    (ulong)keyid[0],(ulong)keyid[1],
-		    datestr_from_sk( sk ),
-		    sk->expiredate? strtimestamp(sk->expiredate):""
+		    colon_datestr_from_sk( sk ),
+		    colon_strtime (sk->expiredate)
 		    /* fixme: add LID */ );
 	}
 	else
@@ -891,7 +909,7 @@ list_node( CTX c, KBNODE node )
 			putchar(':');
 		    putchar('\n');
 		    if( opt.fingerprint && !any )
-			print_fingerprint( NULL, sk );
+			print_fingerprint( NULL, sk, 0 );
 		    any=1;
 		}
 		else if( node->pkt->pkttype == PKT_SECRET_SUBKEY ) {
@@ -906,7 +924,7 @@ list_node( CTX c, KBNODE node )
 	if( !any )
 	    putchar('\n');
 	if( !mainkey && opt.fingerprint > 1 )
-	    print_fingerprint( NULL, sk );
+	    print_fingerprint( NULL, sk, 0 );
     }
     else if( node->pkt->pkttype == PKT_SIGNATURE  ) {
 	PKT_signature *sig = node->pkt->pkt.signature;
@@ -925,7 +943,7 @@ list_node( CTX c, KBNODE node )
 	    fputs("sig", stdout);
 	if( opt.check_sigs ) {
 	    fflush(stdout);
-	    switch( (rc2=do_check_sig( c, node, &is_selfsig )) ) {
+	    switch( (rc2=do_check_sig( c, node, &is_selfsig, NULL )) ) {
 	      case 0:		       sigrc = '!'; break;
 	      case G10ERR_BAD_SIGN:    sigrc = '-'; break;
 	      case G10ERR_NO_PUBKEY: 
@@ -953,7 +971,7 @@ list_node( CTX c, KBNODE node )
 		putchar(sigrc);
 	    printf("::%d:%08lX%08lX:%s::::", sig->pubkey_algo,
 					     (ulong)sig->keyid[0],
-		       (ulong)sig->keyid[1], datestr_from_sig(sig));
+		       (ulong)sig->keyid[1], colon_datestr_from_sig(sig));
 	}
 	else
 	    printf("%c       %08lX %s   ",
@@ -975,7 +993,7 @@ list_node( CTX c, KBNODE node )
 	    m_free(p);
 	}
 	if( opt.with_colons )
-	    printf(":%02x:", sig->sig_class );
+	    printf(":%02x%c:", sig->sig_class, sig->flags.exportable?'x':'l');
 	putchar('\n');
     }
     else
@@ -1174,7 +1192,7 @@ check_sig_and_print( CTX c, KBNODE node )
 {
     PKT_signature *sig = node->pkt->pkt.signature;
     const char *astr, *tstr;
-    int rc;
+    int rc, is_expkey=0;
 
     if( opt.skip_verify ) {
 	log_info(_("signature verification suppressed\n"));
@@ -1191,22 +1209,48 @@ check_sig_and_print( CTX c, KBNODE node )
      * data+sig, sig+data,sig+data and we have not yet encountered the last
      * data, we could also see this a one data with 2 signatures and then 
      * data+sig.
-     * To protect against this we check that we all signatures follow
+     * To protect against this we check that all signatures follow
      * without any intermediate packets.  Note, that we won't get this
      * error when we use onepass packets or cleartext signatures because
      * we reset the list every time
+     *
+     * FIXME: Now that we have these marker packets, we should create a 
+     * real grammar and check against this.
      */
     {
         KBNODE n;
-        int tmp=0;
+        int n_sig=0;
 
-        for(n=c->list; n; n=n->next ) {
-            if ( tmp && n->pkt->pkttype == PKT_SIGNATURE ) {
-                log_error("can't handle these multiple signatures\n");
-                return 0;
+        for (n=c->list; n; n=n->next ) {
+            if ( n->pkt->pkttype == PKT_SIGNATURE ) 
+                n_sig++;
+        }
+        if (n_sig > 1) { /* more than one signature - check sequence */
+            int tmp, onepass;
+
+            for (tmp=onepass=0,n=c->list; n; n=n->next ) {
+                if (n->pkt->pkttype == PKT_ONEPASS_SIG) 
+                    onepass++;
+                else if (n->pkt->pkttype == PKT_GPG_CONTROL
+                         && n->pkt->pkt.gpg_control->control
+                            == CTRLPKT_CLEARSIGN_START ) {
+                    onepass++; /* handle the same way as a onepass */
+                }
+                else if ( (tmp && n->pkt->pkttype != PKT_SIGNATURE) ) {
+                    log_error(_("can't handle these multiple signatures\n"));
+                    return 0;
+                }
+                else if ( n->pkt->pkttype == PKT_SIGNATURE ) 
+                    tmp = 1;
+                else if (!tmp && !onepass 
+                         && n->pkt->pkttype == PKT_GPG_CONTROL
+                         && n->pkt->pkt.gpg_control->control
+                            == CTRLPKT_PLAINTEXT_MARK ) {
+                    /* plaintext before signatures but no one-pass packets*/
+                    log_error(_("can't handle these multiple signatures\n"));
+                    return 0;
+                }
             }
-            else if ( n->pkt->pkttype == PKT_SIGNATURE ) 
-                tmp = 1;
         }
     }
     
@@ -1217,42 +1261,80 @@ check_sig_and_print( CTX c, KBNODE node )
     log_info(_("Signature made %.*s using %s key ID %08lX\n"),
 	    (int)strlen(tstr), tstr, astr? astr: "?", (ulong)sig->keyid[1] );
 
-    rc = do_check_sig(c, node, NULL );
-    if( rc == G10ERR_NO_PUBKEY && opt.keyserver_name && opt.auto_key_retrieve) {
-	if( !hkp_ask_import( sig->keyid ) )
-	    rc = do_check_sig(c, node, NULL );
+    rc = do_check_sig(c, node, NULL, &is_expkey );
+    if( rc == G10ERR_NO_PUBKEY && opt.keyserver_scheme && opt.keyserver_options.auto_key_retrieve) {
+	if( keyserver_import_keyid ( sig->keyid )==0 )
+	    rc = do_check_sig(c, node, NULL, &is_expkey );
     }
     if( !rc || rc == G10ERR_BAD_SIGN ) {
 	KBNODE un, keyblock;
-	char *us;
-	int count=0;
+	int count=0, statno;
+        char keyid_str[50];
+
+	if(rc)
+	  statno=STATUS_BADSIG;
+	else if(sig->flags.expired)
+	  statno=STATUS_EXPSIG;
+	else if(is_expkey)
+	  statno=STATUS_EXPKEYSIG;
+	else
+	  statno=STATUS_GOODSIG;
 
 	keyblock = get_pubkeyblock( sig->keyid );
 
-	us = get_long_user_id_string( sig->keyid );
-	write_status_text( rc? STATUS_BADSIG : STATUS_GOODSIG, us );
-	m_free(us);
+        sprintf (keyid_str, "%08lX%08lX [uncertain] ",
+                 (ulong)sig->keyid[0], (ulong)sig->keyid[1]);
 
-        /* find an print the primary user ID */
+        /* find and print the primary user ID */
 	for( un=keyblock; un; un = un->next ) {
 	    if( un->pkt->pkttype != PKT_USER_ID )
 		continue;
+	    if ( !un->pkt->pkt.user_id->created )
+	        continue;
             if ( un->pkt->pkt.user_id->is_revoked )
                 continue;
             if ( !un->pkt->pkt.user_id->is_primary )
                 continue;
             
+            keyid_str[17] = 0; /* cut off the "[uncertain]" part */
+            write_status_text_and_buffer (statno, keyid_str,
+                                          un->pkt->pkt.user_id->name,
+                                          un->pkt->pkt.user_id->len, 
+                                          -1 );
+
             log_info(rc? _("BAD signature from \"")
-                       : _("Good signature from \""));
+                       : sig->flags.expired ? _("Expired signature from \"")
+		       : _("Good signature from \""));
 	    print_utf8_string( log_stream(), un->pkt->pkt.user_id->name,
 					     un->pkt->pkt.user_id->len );
 	    fputs("\"\n", log_stream() );
             count++;
 	}
 	if( !count ) {	/* just in case that we have no userid */
-	    log_info(rc? _("BAD signature from \"")
+            for( un=keyblock; un; un = un->next ) {
+                if( un->pkt->pkttype == PKT_USER_ID )
+                    break;
+            }
+
+            if (opt.always_trust || !un)
+                keyid_str[17] = 0; /* cut off the "[uncertain]" part */
+
+            write_status_text_and_buffer (statno, keyid_str,
+                                          un? un->pkt->pkt.user_id->name:"[?]",
+                                          un? un->pkt->pkt.user_id->len:3, 
+                                          -1 );
+
+            log_info(rc? _("BAD signature from \"")
+                       : sig->flags.expired ? _("Expired signature from \"")
 		       : _("Good signature from \""));
-	    fputs("[?]\"\n", log_stream() );
+            if (!opt.always_trust && un) {
+                fputs(_("[uncertain]"), log_stream() );
+                putc(' ', log_stream() );
+            }
+            print_utf8_string( log_stream(),
+                               un? un->pkt->pkt.user_id->name:"[?]",
+                               un? un->pkt->pkt.user_id->len:3 );
+	    fputs("\"\n", log_stream() );
 	}
 
         /* If we have a good signature and already printed 
@@ -1273,6 +1355,7 @@ check_sig_and_print( CTX c, KBNODE node )
             }
 	}
 	release_kbnode( keyblock );
+
 	if( !rc )
 	    print_notation_data( sig );
 
@@ -1282,16 +1365,17 @@ check_sig_and_print( CTX c, KBNODE node )
 
 	    if( !get_pubkey( pk, sig->keyid ) ) {
 		byte array[MAX_FINGERPRINT_LEN], *p;
-		char buf[MAX_FINGERPRINT_LEN*2+61];
+		char buf[MAX_FINGERPRINT_LEN*2+72];
 		size_t i, n;
 
 		fingerprint_from_pk( pk, array, &n );
 		p = array;
 		for(i=0; i < n ; i++, p++ )
 		    sprintf(buf+2*i, "%02X", *p );
-		sprintf(buf+strlen(buf), " %s %lu",
+		sprintf(buf+strlen(buf), " %s %lu %lu",
 					 strtimestamp( sig->timestamp ),
-					 (ulong)sig->timestamp );
+					 (ulong)sig->timestamp,
+			                 (ulong)sig->expiredate );
 		write_status_text( STATUS_VALIDSIG, buf );
 	    }
 	    free_public_key( pk );
@@ -1299,6 +1383,15 @@ check_sig_and_print( CTX c, KBNODE node )
 
 	if( !rc )
 	    rc = check_signatures_trust( sig );
+
+	if(sig->flags.expired)
+	  {
+	    log_info("Signature expired %s\n",asctimestamp(sig->expiredate));
+	    rc=G10ERR_GENERAL; /* need a better error here? */
+	  }
+	else if(sig->expiredate)
+	  log_info("Signature expires %s\n",asctimestamp(sig->expiredate));
+
 	if( rc )
 	    g10_errors_seen = 1;
 	if( opt.batch && rc )
@@ -1334,6 +1427,18 @@ proc_tree( CTX c, KBNODE node )
     if( opt.list_packets || opt.list_only )
 	return;
 
+    /* we must skip our special plaintext marker packets here becuase
+       they may be the root packet.  These packets are only used in
+       addionla checks and skipping them here doesn't matter */
+    while ( node
+            && node->pkt->pkttype == PKT_GPG_CONTROL
+            && node->pkt->pkt.gpg_control->control
+                         == CTRLPKT_PLAINTEXT_MARK ) {
+        node = node->next;
+    }
+    if (!node)
+        return;
+
     c->local_id = 0;
     c->trustletter = ' ';
     if( node->pkt->pkttype == PKT_PUBLIC_KEY
@@ -1364,7 +1469,7 @@ proc_tree( CTX c, KBNODE node )
 	    }
 	    else {
 		rc = ask_for_detached_datafile( c->mfx.md, c->mfx.md2,
-						iobuf_get_fname(c->iobuf),
+						iobuf_get_real_fname(c->iobuf),
 			n1? (n1->pkt->pkt.onepass_sig->sig_class == 0x01):0 );
 	    }
 	    if( rc ) {
@@ -1442,7 +1547,7 @@ proc_tree( CTX c, KBNODE node )
 	    }
 	    else {
 		rc = ask_for_detached_datafile( c->mfx.md, c->mfx.md2,
-						iobuf_get_fname(c->iobuf),
+						iobuf_get_real_fname(c->iobuf),
 						(sig->sig_class == 0x01) );
 	    }
 	    if( rc ) {
@@ -1454,14 +1559,19 @@ proc_tree( CTX c, KBNODE node )
             log_error (_("not a detached signature\n") );
             return;
         }
-	else
+        else if ( c->pipemode.op == 'B' )
+            ; /* this is a detached signature trough the pipemode handler */
+	else if (!opt.quiet)
 	    log_info(_("old style (PGP 2.x) signature\n"));
 
-	check_sig_and_print( c, node );
+	for( n1 = node; n1; (n1 = find_next_kbnode(n1, PKT_SIGNATURE )) )
+	    check_sig_and_print( c, n1 );
     }
-    else
+    else {
+        dump_kbnode (c->list);
 	log_error(_("invalid root packet detected in proc_tree()\n"));
-
+        dump_kbnode (node);
+    }
 }
 
 
